#########################################################
# March, 2017
# Kan Li
# 
# Online supplement for "Functional Joint Model for Longitudinal 
# and Time-to-Event Data: An Application to Alzheimer's Disease"
#
# Main functions for fitting FJM model on simulated data 
# (Setting 1) using author's code (myFJM) for fitting FPCA and JM. 
# 
#########################################################

rm(list=ls())
source("./myFJM.R")


###############################################################
## set simulation design elements
###############################################################
D <- 10 ## size of grid for functional data
by <- 0.1
s <- seq(0, D, by=by)  # Domain S
I <- 200  ## sample size
J <- 4   ## number of visits
beta0 <- 1.35
beta1 <- 0.78
var.b <- 1  ## variance for random effect
var.eps <- 1 ## variance for error
gamma1 <- -1.75
alpha <- 0.29
trueBeta <- 2*sin(s*pi/5)  # B^x(s)
trueGamma <- 1.2*sin(s*pi/4) # B^w(s)



###############################################################
## start main function
###############################################################
# load simulated data
load('./simulated_data.rdata')
N.sim <- 10
coefficients <- Betas <- Gammas <- NULL
for(run in 1: N.sim){
  long = sim.data[[run]]$data.long
  surv = sim.data[[run]]$data.surv
  
  # process the functional predictor
  FPCA.X <- process.funcs(long$ID, long$func.X, method="knots", smooth.cov=T)  # Select number of FPC based on "knots" to insure constant in each simulation run.
  FPCA.W <- process.funcs(surv$ID, surv$func.W, method="knots", smooth.cov=T)
  long$FPCscore.X <- I(FPCA.X$FPCscore)
  surv$FPCscore.W <- I(FPCA.W$FPCscore)
  
  ctrl = lmeControl(maxIter=3000, msMaxIter=3000, niterEM=3000, msMaxEval=3000,returnObject=TRUE, opt = c("optim")) 
  # fit joint model
  fit.long <- lme(Y ~ obstime + FPCscore.X, random= ~1|ID, data=long, control=ctrl)
  fit.surv <- coxph(Surv(time, event)~ W1 + FPCscore.W, data = surv, x=T, control=coxph.control(iter.max=50))
  
  fit.joint <- try(myFJM(fit.long, fit.surv, timeVar = "obstime", method = "weibull-PH-GH"))
  
  if (inherits(fit.joint, "try-error")) {
    print("error")
    break
  }
  
  # estimated coefficient functions
  Betas <- rbind(Betas, c(FPCA.X$basis%*%(fit.joint$coefficients$betas[-c(1:2)])))
  Gammas <- rbind(Gammas, c(FPCA.W$basis%*%(fit.joint$coefficients$gammas[-c(1:2)])))
  coefficients <- rbind(coefficients, c(fit.joint$coefficients$betas[2], fit.joint$coefficients$D, 
                                        fit.joint$coefficients$sigma, fit.joint$coefficients$gammas[2], 
                                        fit.joint$coefficients$alpha))
  
}

###############################################################
## end model fitting
###############################################################

# summary coefficients
BetaHat <- apply(Betas,2,function(x) mean(x, na.rm = T))
GammaHat <- apply(Gammas,2,function(x) mean(x, na.rm = T))
coeffHat <- apply(coefficients, 2, function(x) mean(x, na.rm=T))

# plot functional coefficient
par(mfrow=c(1,2))
plot(s, BetaHat, type = "l", xlab="s", ylab= expression(beta(s)), lwd = 2)
lines(s,trueBeta, col="red", lwd = 2)

plot(s, GammaHat, type = "l", xlab="s", ylab= expression(gamma(s)), lwd = 2)
lines(s,trueGamma, col="red", lwd = 2)

# print estimations
estimations <- rbind(c(beta1, var.b, var.eps, gamma1, alpha), coeffHat)
rownames(estimations) <- c("true", "estimates")
colnames(estimations) <- c("beta1", "var.b", "var.eps", "gamma1", "alpha")
print(estimations, digits = 3)
