#########################################################
# March, 2017
# Kan Li
# 
# Online supplement for "Functional Joint Model for Longitudinal 
# and Time-to-Event Data: An Application to Alzheimer's Disease"
#
# Main functions for fitting FJM model on simulated data 
# (Setting 1) using available packages.
# 
#########################################################

rm(list=ls())
library(refund)
library(JM)


###############################################################
## set simulation design elements
###############################################################
D <- 10 ## size of grid for functional data
by <- 0.1
s <- seq(0, D, by=by)  # Domain S
I <- 200  ## sample size
J <- 4   ## number of visits
beta0 <- 1.35
beta1 <- 0.78
var.b <- 1  ## variance for random effect
var.eps <- 1 ## variance for error
gamma1 <- -1.75
alpha <- 0.29
trueBeta <- 2*sin(s*pi/5)  # B^x(s)
trueGamma <- 1.2*sin(s*pi/4) # B^w(s)


###############################################################
## define support functions 
###############################################################

# FPCA: calculate FPC score
process.funcs = function(subjs, funcs, method=c("pve","knots"),pve = 0.95, knots=5){
  ids <- as.numeric(as.factor(subjs))
  baseline.funcs <- as.matrix(funcs[!duplicated(ids),])
  if(method=="pve"){
    fpca <- fpca.sc(Y=baseline.funcs, pve=pve, var=T) 
  }else{
    fpca <- fpca.sc(Y=baseline.funcs, npc=knots, var=T)
  }
  fpcscore <- fpca$scores[ids, ]
  kz <- dim(fpcscore)[2]
  phi <- fpca$efunctions[,1:kz]
  output <- list(FPCscore = fpcscore, kz = kz, basis = phi)
  return(output)
}

# Fit FJM model
FJM = function(long, surv){
  
  # process the functional predictor
  FPCA.X <- process.funcs(long$ID, long$func.X, method="knots")  # Select number of FPC based on "knots" to insure constant in each simulation run.
  FPCA.W <- process.funcs(surv$ID, surv$func.W, method="knots")
  long$FPCscore.X <- I(FPCA.X$FPCscore)
  surv$FPCscore.W <- I(FPCA.W$FPCscore)
  
  ctrl = lmeControl(maxIter=3000, msMaxIter=3000, niterEM=3000, msMaxEval=3000,returnObject=TRUE, opt = c("optim")) 
  # fit joint model
  fit.long <- lme(Y ~ obstime + FPCscore.X, random= ~1|ID, data=long, control=ctrl)
  fit.surv <- coxph(Surv(time, event)~ W1 + FPCscore.W, data = surv, x=T, control=coxph.control(iter.max=50))
  
  fit.joint <- try(jointModel(fit.long, fit.surv, timeVar = "obstime", method = "weibull-PH-GH"))
  
  if (inherits(fit.joint, "try-error")) {
    print("error")
    break
  }
  
  # estimated coefficient function
  BetaHat <- c(FPCA.X$basis%*%(fit.joint$coefficients$betas[-c(1:2)])*by)
  GammaHat <- c(FPCA.W$basis%*%(fit.joint$coefficients$gammas[-c(1:2)])*by)
  
  output <- list(FPCA.X = FPCA.X, FPCA.W = FPCA.W, FJM = fit.joint, BetaHat=BetaHat, GammaHat=GammaHat)
  return(output)
}


###############################################################
## start main function
###############################################################
# load simulated data
load('./simulated_data.rdata')
N.sim <- 10
coefficients <- Betas <- Gammas <- NULL
for(run in 1: N.sim){
  tmp.data <- sim.data[[run]]
  fit.joint <- FJM(tmp.data$data.long, tmp.data$data.surv)
  Betas <- rbind(Betas, fit.joint$BetaHat)
  Gammas <- rbind(Gammas, fit.joint$GammaHat)
  coefficients <- rbind(coefficients, c(fit.joint$FJM$coefficients$betas[2], fit.joint$FJM$coefficients$D, 
                                        fit.joint$FJM$coefficients$sigma, fit.joint$FJM$coefficients$gammas[2], 
                                        fit.joint$FJM$coefficients$alpha))
}

###############################################################
## end model fitting
###############################################################

# summary coefficients
BetaHat <- apply(Betas,2,function(x) mean(x, na.rm = T))
GammaHat <- apply(Gammas,2,function(x) mean(x, na.rm = T))
coeffHat <- apply(coefficients, 2, function(x) mean(x, na.rm=T))

# plot functional coefficient
par(mfrow=c(1,2))
plot(s, BetaHat, type = "l", xlab="s", ylab= expression(beta(s)), lwd = 2)
lines(s,trueBeta, col="red", lwd = 2)

plot(s, GammaHat, type = "l", xlab="s", ylab= expression(gamma(s)), lwd = 2)
lines(s,trueGamma, col="red", lwd = 2)

# print estimations
estimations <- rbind(c(beta1, var.b, var.eps, gamma1, alpha), coeffHat)
rownames(estimations) <- c("true", "estimates")
colnames(estimations) <- c("beta1", "var.b", "var.eps", "gamma1", "alpha")
print(estimations, digits = 3)



