##################################################################################
# 11/16/2015  Kan Li
# This code is used to fit functional joint model via EM algorithm
# It supports Weibull and Piecewise base hazard function
# Model formulation with only m(t)
# Borrowed and modified core functions from JM package (Rizopoulos, 2010)
##################################################################################

library(nlme)
library(survival)
library(mgcv)
library(MASS)
source("./myFJMSupport.R")

process.funcs = function(subjs, funcs, method=c("pve","knots"),pve = 0.95, knots=5, smooth.cov=T){
  ids <- as.numeric(as.factor(subjs))
  baseline.funcs <- as.matrix(funcs[!duplicated(ids),])
  D <- dim(funcs)[2]
  
  mean.funcs <- apply(baseline.funcs, 2, mean, na.rm=TRUE)
  resd <- sapply(1:D, function(u) baseline.funcs[,u]- mean.funcs[u])
  resd.o <- resd
  
  # construct and smooth covariance matrices
  G.sum <- matrix(0, D, D)
  G.count <- matrix(0, D, D)
  
  for(i in 1:dim(resd)[1]){
    row.ind <-i
    temp <- resd[row.ind, ] %*% t( resd[row.ind, ])
    G.sum <- G.sum + replace(temp, which(is.na(temp)), 0)
    G.count <- G.count + as.numeric(!is.na(temp))
  }
  G <- ifelse(G.count==0, NA,  G.sum/G.count)   
  if(smooth.cov) {
    G2 <- G
    diag(G2) = rep(NA, D)
    g2 <- as.vector(G2)
    N <- 10
    t1 <- rep(s, D)
    t2 <- rep(s, each=D)  
    newdata <- data.frame(t1 = t1, t2 = t2)
    K.0 <- matrix(predict(gam(as.vector(g2) ~ te(t1,t2, k = N)), newdata), D, D)
    K.0 <- (K.0 + t(K.0))/2
    eigenDecomp <- eigen(K.0)
  }
  else{
    eigenDecomp <- eigen(G)
  }
  
  eigenDecomp$vectors = eigenDecomp$vectors/sqrt(by)
  
  if(method=="pve"){
    cum.pve <- cumsum(eigenDecomp$values)/sum(values)
    knots <- which(cum.pve>pve)[1]
  }
  
  phi=eigenDecomp$vectors[,1:knots]
  
  scores = matrix(0, dim(resd.o)[1], knots) 
  for(i in 1:dim(resd.o)[1])   {
    for(j in 1:knots) {
      scores[ i ,j] <- sum(resd.o[i,] * phi[,j] ) * by   
    }
  }
  
  fpcscore <- scores[ids, ]
  kz <- knots
  output <- list(FPCscore = fpcscore, kz = kz, basis = phi)
  return(output)
}
  
myFJM = function(lmeObject, survObject,timeVar, parameterization = "value",
                method = c("weibull-PH-GH", "piecewise-PH-GH"),  control = list(), ...){
  
  #-----------------------------------------------------------------
  # processing survival object
  formT <- formula(survObject)  # formular of survobject 
  W <- survObject$x             # baseline covariate
  keepW <- suppressWarnings(!is.na(survObject$coefficients))
  W <- W[, keepW, drop = FALSE]
  nRisks <- 1  
  surv <- survObject$y          # surv time and censor information
  if (attr(surv, "type") == "right") {
    LongFormat <- FALSE
    Time <- survObject$y[, 1]  
    d <- survObject$y[, 2]
  }
  idT = seq_along(Time)
  idT <- match(idT, unique(idT))
  nT <- length(unique(idT))
  WintF.vl <- WintF.sl <- as.matrix(rep(1, length(Time)))
  
  # processing longitudinal object
  id <- as.vector(unclass(lmeObject$groups[[1]]))
  b <- data.matrix(ranef(lmeObject))
  dimnames(b) <- NULL
  nY <- nrow(b)
  if (nY != nT) 
    stop("sample sizes in the longitudinal and event processes differ; ")
  TermsX <- lmeObject$terms 
  data <- lmeObject$data[all.vars(TermsX)]
  data <- data[complete.cases(data), ]
  formYx <- formula(lmeObject)
  mfX <- model.frame(TermsX, data = data)
  X <- model.matrix(formYx, mfX)
  
  formYz <- formula(lmeObject$modelStruct$reStruct[[1]]) # random effect information
  mfZ <- model.frame(terms(formYz), data = data)
  TermsZ <- attr(mfZ, "terms")
  Z <- model.matrix(formYz, mfZ)
  y.long <- model.response(mfX, "numeric")  # prepare y long form
  data.id <- data[!duplicated(id), ]
  data.id <- data.id[idT, ]    # assure consistency with the number of subjects in cox model
  
  # controls 
  interFact = NULL; derivForm = NULL; lag = 0; scaleWB = NULL; CompRisk = FALSE; init = NULL
  ########################################################################
  
  if (!timeVar %in% names(data)) 
    stop("\n'timeVar' does not correspond to one of the columns in the model.frame of 'lmeObject'.")
  max.timeY <- tapply(data[[timeVar]], id, max)
  max.timeT <- tapply(Time, idT, max)
  if (!all(max.timeT >= max.timeY)) {
    stop("\nit seems that there are longitudinal measurements taken after the event times for some subjects ).")
  }
  
  data.id[[timeVar]] <- pmax(Time - lag, 0)
  
  if (parameterization == "value") {
    mfX.id <- model.frame(TermsX, data = data.id)
    mfZ.id <- model.frame(TermsZ, data = data.id) # ? form of this
    Xtime <- model.matrix(formYx, mfX.id)  # ?
    Ztime <- model.matrix(formYz, mfZ.id)
    long <- c(X %*% fixef(lmeObject)) + rowSums(Z * b[id, ])
    long.deriv <- NULL
  } # create diff form of data with baseline measures but survival time
  
  y <- list(y = y.long, logT = log(Time), d = d, lag = lag) # data for survival part
  x <- list(X = X, Z = Z, W = W, WintF.vl = WintF.vl, WintF.sl = WintF.sl, 
            idT = idT, nRisks = nRisks) # data for longitudinal part
  x <- c(x, list(Xtime = Xtime, Ztime = Ztime)) # add baseline data
  ind.noadapt = T
  
  # model fitting constrain
  #----------------------------------------------------------------------------
  con <- list(only.EM = FALSE, iter.EM =  50, 
              iter.qN = 350, optimizer = "optim", tol1 = 0.001, tol2 = 1e-04, 
              tol3 = if (!CompRisk) sqrt(.Machine$double.eps) else 1e-09, 
              numeriDeriv = "fd", eps.Hes = 1e-06, parscale = NULL, 
              step.max = 0.1, backtrackSteps = 2, knots = NULL, ObsTimes.knots = TRUE, 
              lng.in.kn = if (method == "piecewise-PH-GH") 6 else 5,
              ord = 4, equal.strata.knots = TRUE, typeGH = if (ind.noadapt) "simple" else "adaptive", 
              GHk = if (ncol(Z) < 3 && nrow(Z) < 2000) 15 else 9, GKk = if (method == "piecewise-PH-GH" || length(Time) > nRisks * nT) 7 else 15, 
              verbose = FALSE)
  
  control <- c(control)
  namC <- names(con) # control items
  con[(namc <- names(control))] <- control
  
  # prepare cox model integration
  #--------------------------------------------------------------------
  if (method  == "weibull-PH-GH") {
    wk <- gaussKronrod(con$GKk)$wk
    sk <- gaussKronrod(con$GKk)$sk
    if (LongFormat) {
      Time0 <- survObject$y[, 1]
      P <- (Time - Time0)/2
      P1 <- (Time + Time0)/2
      st <- outer(P, sk) + P1
    }
    else {
      P <- as.vector(Time)/2
      st <- outer(P, sk + 1)
    }
    dimnames(st) <- names(P) <- NULL
    id.GK <- rep(seq_along(Time), each = con$GKk)
    data.id2 <- data.id[id.GK, , drop = FALSE]
    data.id2[[timeVar]] <- pmax(c(t(st)) - lag, 0)
    if (parameterization %in% c("value")) {
      mfX <- model.frame(TermsX, data = data.id2)
      mfZ <- model.frame(TermsZ, data = data.id2)
      Xs <- model.matrix(formYx, mfX)
      Zs <- model.matrix(formYz, mfZ)
    }
    
    Ws.intF.vl <- WintF.vl[id.GK, , drop = FALSE]
    Ws.intF.sl <- WintF.sl[id.GK, , drop = FALSE]
    x <- c(x, list(P = P, st = c(t(st)), wk = wk, Ws.intF.vl = Ws.intF.vl, 
                   Ws.intF.sl = Ws.intF.sl))
    x <- c(x, list(Xs = Xs, Zs = Zs))
  }
  
  if (method == "piecewise-PH-GH") {
    wk <- gaussKronrod(con$GKk)$wk
    sk <- gaussKronrod(con$GKk)$sk
    nk <- length(sk)
    if (is.null(con$knots) || !is.numeric(con$knots)) {
      Q <- con$lng.in.kn + 1
      qs <- if (con$ObsTimes.knots) {
        unique(quantile(Time, seq(0, 1, len = Q + 1), 
                        names = FALSE)[-c(1, Q + 1)])
      }
      qs <- qs + 1e-06
      if (max(qs) > max(Time)) 
        qs[which.max(qs)] <- max(Time) - 1e-06
      con$knots <- qs
      qs <- c(0, qs, max(Time) + 1)
      Q <- length(qs) - 1
    }
    ind <- findInterval(Time, qs, rightmost.closed = TRUE)
    
    D <- matrix(0, length(ind), Q) 
    D[cbind(seq_along(ind), ind)] <- 1
    D <- D * d # d cencor index
    Tiq <- outer(Time, qs, pmin)
    Lo <- Tiq[, 1:Q]
    Up <- Tiq[, 2:(Q + 1)]
    T <- Up - Lo
    P <- T/2 # find mid time point
    P[P < con$tol3] <- as.numeric(NA)
    P1 <- (Up + Lo)/2
    st <- matrix(0, nY, nk * Q)
    skQ <- rep(sk, Q)
    for (i in seq_len(nY)) {
      st[i, ] <- rep(P[i, ], each = nk) * skQ + rep(P1[i, ], each = nk)
    }
    y <- c(y, list(ind.D = ind))
    id.GK <- rep(seq_len(nY), rowSums(!is.na(st)))
    P <- c(t(P))
    data.id2 <- data.id[rep(seq_len(nY), each = nk * Q), ]
    data.id2[[timeVar]] <- pmax(c(t(st)) - lag, 0)
    data.id2 <- data.id2[!is.na(data.id2[[timeVar]]), ]
    if (parameterization %in% c("value")) {
      mfX <- model.frame(TermsX, data = data.id2)
      mfZ <- model.frame(TermsZ, data = data.id2)
      Xs <- model.matrix(formYx, mfX)
      Zs <- model.matrix(formYz, mfZ)
    }
    
    Ws.intF.vl <- WintF.vl[id.GK, , drop = FALSE]
    Ws.intF.sl <- WintF.sl[id.GK, , drop = FALSE]
    x <- c(x, list(P = P[!is.na(P)], st = st[!is.na(st)], 
                   wk = wk, id.GK = id.GK, Q = Q, Ws.intF.vl = Ws.intF.vl, 
                   Ws.intF.sl = Ws.intF.sl))
    x <- switch(parameterization, value = c(x, list(Xs = Xs, Zs = Zs)), 
                slope = c(x, list(Xs.deriv = Xs.deriv, Zs.deriv = Zs.deriv)), 
                both = c(x, list(Xs.deriv = Xs.deriv, Zs.deriv = Zs.deriv, Xs = Xs, Zs = Zs)))
  }
  #--------------------------------------------------------------------
  
  ####################################################
  VC <- lapply(pdMatrix(lmeObject$modelStruct$reStruct), "*", lmeObject$sigma^2)[[1]]
  con$inv.chol.VC <- solve(chol(solve(VC)))
  con$det.inv.chol.VC <- det(con$inv.chol.VC)
  con$ranef <- b
  if (all(VC[upper.tri(VC)] == 0)) 
    VC <- diag(VC)
  ######################################################
  
  init.surv <- initial.surv(Time, d, W, WintF.vl, WintF.sl, 
                            id, times = data[[timeVar]], method, parameterization, 
                            long = long,  extra = list(W2 = x$W2,control = con, ii = idT, strata = survObject$strata), 
                            LongFormat = CompRisk)
  
  initial.values <- c(list(betas = fixef(lmeObject), sigma = lmeObject$sigma, 
                           D = VC), init.surv)
  
  if(method == "weibull-PH-GH"){
    out = weibullPHGH.fit(x, y, id, initial.values,  scaleWB, parameterization, derivForm, con)
  }else if( method == "piecewise-PH-GH"){
    out = piecewisePHGH.fit(x, y, id, initial.values, parameterization, derivForm, con)
  }
  
  #-----------------
  H <- out$Hessian
  if (any(is.na(H) | !is.finite(H))) {
    warning("infinite or missing values in Hessian at convergence.\n")
  }
  else {
    ev <- eigen(H, symmetric = TRUE, only.values = TRUE)$values
    if (!all(ev >= -1e-06 * abs(ev[1]))) 
      warning("Hessian matrix at convergence is not positive definite.\n")
  }
  out$coefficients <- out$coefficients[!sapply(out$coefficients, 
                                               is.null)]
  out$x <- x
  out$y <- y
  out$times <- data[[timeVar]]
  out$data <- data
  out$data.id <- data.id
  out$method <- method
  out$termsYx <- TermsX
  out$termsYz <- TermsZ
  out$termsT <- survObject$terms
  out$formYx <- formYx
  out$formYz <- formYz
  out$formT <- formT
  out$timeVar <- timeVar
  out$control <- con
  out$parameterization <- parameterization
  out$derivForm <- derivForm
  out$interFact <- interFact
  out$CompRisk <- CompRisk
  out$LongFormat <- LongFormat
  out$assignY <- attr(lmeObject$fixDF, "assign")[-1]
  out$assignT <- survObject$assign
  class(out) <- "FJM"
  out
}